#!/usr/bin/env python

"""
This program is used to create example NeXus files using metadata from an input .json file.

Example:
    python writer.py output_nexus.h5 input_metadata.json
    python writer.py output_nexus.nxs input_metadata.json

A custom .json file containing metadata can be created by modifying and running create_metadata.py.
"""
import os
import numpy
import h5py

from . import imgcif2mcstas, create_attributes, set_dependency


def generate_image_data(shape, filename):
    data = numpy.ndarray(shape)
    data.fill(0)
    with h5py.File(filename, "w") as datafile:
        datafile.create_dataset("data", data=data)
    return data


# TODO make vds and add link
def generate_event_data(num_events, outfile):
    outfile.create_dataset("cue_id", data=numpy.zeros(100))
    outfile.create_dataset("cue_timestamp_zero", data=numpy.zeros(100))
    outfile.create_dataset("event_id", data=numpy.zeros(num_events))
    outfile.create_dataset("event_time_offset", data=numpy.zeros(num_events))
    outfile.create_dataset("event_energy", data=numpy.zeros(num_events))


class NexusWriter:
    """
    Class to write metadata in NeXus file.
    """

    def __init__(self, nxsfile: h5py.File, params):
        self._params = params
        # Define coordinate frame
        self._cf = params.input.coordinate_frame

        # Output parameters
        self._nxs = nxsfile
        self._datafile = params.output.data_file_template

        # Unpack input parameters
        self.goniometer = params.input.goniometer
        self.detector = params.input.detector
        self.attenuator = params.input.attenuator
        self.beam = params.input.beam
        self.beamline = params.input.beamline

        # Which kind of data is being generated?
        self._mode = self.detector.mode

    @staticmethod
    def split_arrays(coord_frame, _axes, _array):
        # If coordinate frame is imgcif conver and make a list of arrays
        # else just split into a list
        v = []
        for j in range(len(_axes)):
            a = _array[3 * j : 3 * j + 3]
            if coord_frame == "imgcif":
                v.append(imgcif2mcstas(a))
                # print("Converting to mcstas coordinates ...")
            else:
                v.append(a)
        return v

    @staticmethod
    def calculate_module_offset(
        beam_center_xy, xy_pixel_size, fast_axis_v, slow_axis_v
    ):
        # assumes fast and slow axis vectors have already been converted if needed
        # Scaled center
        x_scaled = beam_center_xy[0] * xy_pixel_size[0]
        y_scaled = beam_center_xy[1] * xy_pixel_size[1]
        # Detector origin
        det_origin = x_scaled * numpy.array(fast_axis_v) + y_scaled * numpy.array(
            slow_axis_v
        )
        det_origin = list(-det_origin)
        return det_origin

    def write_NXdata(self, nxentry):
        # Find scan axis
        goniometer = self.goniometer
        # Scan axis is the on that has a different start and end value
        # idx = numpy.where(goniometer.starts != goniometer.ends)[0][0]
        # TOFIX for some reason this returns the wrong index
        idx = [(i != j) for i, j in zip(goniometer.starts, goniometer.ends)].index(
            True
        )  # TEMPORARY FIX

        nxdata = nxentry.create_group("data")
        create_attributes(
            nxdata,
            ("NX_class", "axes", "signal"),
            ("NXdata", goniometer.axes[idx], "data"),
        )

        # If mode = images, create blank image data, else go to events
        if self._mode == "images":
            _scan_range = numpy.arange(
                goniometer.starts[idx], goniometer.ends[idx], goniometer.increments[idx]
            )
            dset_shape = (self._params.input.n_images,) + tuple(
                self.detector.image_size
            )
            img = generate_image_data(dset_shape, self._datafile)
            nxdata.create_dataset("data", data=img)
            # nxdata["datafile"] = h5py.ExternalLink(self._datafile, "/")
        else:
            # TODO This needs some serious rethining at some point
            dirname = os.path.dirname(self._nxs.filename)
            outfile = os.path.join(dirname, "File_00000{}.h5")
            _scan_range = (goniometer.starts[idx], goniometer.ends[idx])
            # Let's make 2 files of event data
            for n in range(1, 3):
                with h5py.File(outfile.format(n), "w") as out:
                    generate_event_data(self._params.input.n_events, out)
                    nxdata["File_00000{}".format(n)] = h5py.ExternalLink(
                        out.filename, "/"
                    )

        # TODO handle stills
        ax = nxdata.create_dataset(goniometer.axes[idx], data=_scan_range)
        _dep = set_dependency(
            goniometer.depends[idx], path="/entry/sample/transformations/"
        )
        # This could probably be handled better but for the moment it works
        # if self._cf == "imgcif":
        #    _vector = imgcif2mcstas(goniometer.vector[idx])
        # else:
        #    _vector = goniometer.vectors[idx]
        vectors = NexusWriter.split_arrays(
            self._cf, goniometer.axes, goniometer.vectors
        )
        create_attributes(
            ax,
            ("depends_on", "transformation_type", "units", "vector"),
            (
                _dep,
                goniometer.types[idx],
                goniometer.units[idx],
                vectors[idx],
                # _vector,
            ),
        )

    def write_NXdetector(self, nxinstr):
        detector = self.detector
        nxdet = nxinstr.create_group("detector")
        create_attributes(nxdet, ("NX_class",), ("NXdetector",))
        nxdet.create_dataset(
            "depends_on", data="/entry/instrument/transformations/det_z"
        )
        # Detector description
        nxdet.create_dataset("description", data=numpy.string_(detector.description))
        nxdet.create_dataset("type", data=numpy.string_("Pixel"))

        # Beam center
        beam_center_x = nxdet.create_dataset(
            "beam_center_x", data=detector.beam_center[0]
        )
        create_attributes(beam_center_x, ("units",), ("pixels",))
        beam_center_y = nxdet.create_dataset(
            "beam_center_y", data=detector.beam_center[1]
        )
        create_attributes(beam_center_y, ("units",), ("pixels",))

        # Pixel size
        x_pix_size = nxdet.create_dataset(
            "x_pixel_size", data=detector.pixel_size[0] / 1000
        )
        create_attributes(x_pix_size, ("units",), ("m",))
        y_pix_size = nxdet.create_dataset(
            "y_pixel_size", data=detector.pixel_size[1] / 1000
        )
        create_attributes(y_pix_size, ("units",), ("m",))

        # Count time
        nxdet.create_dataset("count_time", data=detector.exposure_time)

        # Sensor material, sensor thickness
        nxdet.create_dataset(
            "sensor_material", data=numpy.string_(detector.sensor_material)
        )
        nxdet.create_dataset("sensor_thickness", data=detector.sensor_thickness / 1000)
        create_attributes(nxdet["sensor_thickness"], ("units",), ("m",))

        if self._mode == "images":
            # Overload and underload
            nxdet.create_dataset("saturation_value", data=detector.overload)
            nxdet.create_dataset("underload_value", data=detector.underload)

        # Flatfield
        if detector.flatfield:
            nxdet.create_dataset("flatfield_applied", data=False)
            nxdet.create_dataset(
                "flatfield", data="ExternalLink/to/flatfield/correction/data"
            )
        # Bad pixel mask
        if detector.pixel_mask:
            nxdet.create_dataset("pixel_mask_applied", data=False)
            nxdet.create_dataset(
                "pixel_mask", data="ExternalLink/to/mask/correction/data"
            )

        # Module: /entry/instrument/detector/module
        self.write_NXdetector_module(nxdet)

        # DetectorSpecific: /entry/instrument/detector/detectorSpecific
        self.write_NXdetectorSpec(nxdet)

    def write_NXdetector_module(self, nxdet):
        detector = self.detector
        nxmod = nxdet.create_group("module")
        create_attributes(nxmod, ("NX_class",), ("NXdetector_module",))
        nxmod.create_dataset("data_origin", data=numpy.array([0, 0]))
        nxmod.create_dataset("data_size", data=detector.image_size)
        nxmod.create_dataset("data_stride", data=numpy.array([1, 1]))

        if self._cf == "imgcif":
            fast_axis = imgcif2mcstas(detector.fast_axis)
            slow_axis = imgcif2mcstas(detector.slow_axis)
        else:
            fast_axis = detector.fast_axis
            slow_axis = detector.slow_axis

        offsets = NexusWriter.split_arrays(self._cf, ["fast", "slow"], detector.offsets)
        # TODO this needs some rethinking

        # Fast/slow pixel direction
        fast_pixel = nxmod.create_dataset(
            "fast_pixel_direction", data=detector.pixel_size[0] / 1000
        )
        create_attributes(
            fast_pixel,
            ("depends_on", "offset", "transformation_type", "units", "vector"),
            (
                "/entry/instrument/transformations/det_z",
                offsets[0],
                "translation",
                "m",
                fast_axis,
            ),
        )
        slow_pixel = nxmod.create_dataset(
            "slow_pixel_direction", data=detector.pixel_size[1] / 1000
        )
        create_attributes(
            slow_pixel,
            ("depends_on", "offset", "transformation_type", "units", "vector"),
            (
                "/entry/instrument/detector/module/fast_pixel_direction",
                offsets[1],
                "translation",
                "m",
                slow_axis,
            ),
        )
        # Module_offset
        if detector.module_offset is True:
            offset = NexusWriter.calculate_module_offset(
                detector.beam_center, detector.pixel_size, fast_axis, slow_axis
            )
            module_offset = nxmod.create_dataset("module_offset", data=([0.0]))
            create_attributes(
                module_offset,
                ("depends_on", "offset", "transformation_type", "units", "vector"),
                (
                    "/entry/instrument/transformations/det_z",
                    offset,
                    "translation",
                    "m",
                    [1, 0, 0],
                ),
            )
            # Set correct path for deendency of fast and slow axis
            _path = "/entry/instrument/detector/module/module_offset"
            create_attributes(fast_pixel, ("depends_on"), (_path,))
            create_attributes(slow_pixel, ("depends_on"), (_path,))

    def write_NXdetectorSpec(self, nxdet):
        detector = self.detector
        grp = nxdet.create_group("detectorSpecific")
        grp.create_dataset("x_pixels", data=detector.image_size[0])
        grp.create_dataset("y_pixels", data=detector.image_size[1])
        if self._mode == "images":
            grp.create_dataset("nimages", data=self._params.input.n_images)

    def write_NXpositioner(self, nxinstr):
        detector = self.detector
        vectors = NexusWriter.split_arrays(self._cf, detector.axes, detector.vectors)

        # Detector_z: /entry/instrument/detector_z
        assert "det_z" in detector.axes, "detector_z not in metadata"
        idx = detector.axes.index("det_z")
        nxdet_z = nxinstr.create_group("detector_z")
        create_attributes(nxdet_z, ("NX_class",), ("NXpositioner",))
        det_z = nxdet_z.create_dataset(detector.axes[idx], data=detector.starts[idx])
        _dep = set_dependency(
            detector.depends[idx], "/entry/instrument/transformations/"
        )
        create_attributes(
            det_z,
            ("depends_on", "transformation_type", "units", "vector"),
            (
                _dep,
                detector.types[idx],
                detector.units[idx],
                vectors[idx],
            ),
        )

        # Transformations: /entry/instrument/transformations
        nxtransf = nxinstr.create_group("transformations")
        create_attributes(nxtransf, ("NX_class",), ("NXtransformations",))
        _link = "/entry/instrument/detector_z/" + detector.axes[idx]
        nxtransf["det_z"] = self._nxs[_link]

        # Add two_theta arm
        if "two_theta" in detector.axes:
            i = detector.axes.index("two_theta")
            twotheta = nxtransf.create_dataset(
                detector.axes[i], data=detector.starts[i]
            )
            _dep = set_dependency(
                detector.depends[i],
                "/entry/instrument/transformations/",
            )
            create_attributes(
                twotheta,
                ("depends_on", "transformation_type", "units", "vector"),
                (
                    _dep,
                    detector.types[i],
                    detector.units[i],
                    vectors[i],
                ),
            )

            nx2theta = nxinstr.create_group("twotheta")
            create_attributes(nx2theta, ("NX_class",), ("NXpositioner",))
            nx2theta["two_theta"] = self._nxs[
                "/entry/instrument/transformations/two_theta"
            ]

    def write_NXinstrument(self, nxentry):
        nxinstr = nxentry.create_group("instrument")
        create_attributes(
            nxinstr, ("NX_class", "short_name"), ("NXinstrument", "beamline")
        )

        # NXattenuator: /entry/instrument/attenuator
        attenuator = self.attenuator
        nxatt = nxinstr.create_group("attenuator")
        create_attributes(nxatt, ("NX_class",), ("NXattenuator",))
        nxatt.create_dataset(
            "attenuator_transmission",
            data=attenuator.transmission,
        )

        # NXbeam: /entry/instrument/beam
        beam = self.beam
        nxbeam = nxinstr.create_group("beam")
        create_attributes(nxbeam, ("NX_class",), ("NXbeam",))
        wl = nxbeam.create_dataset("incident_wavelength", data=beam.wavelength)
        create_attributes(wl, ("units",), ("angstrom",))
        flux = nxbeam.create_dataset("total_flux", data=beam.flux)
        create_attributes(flux, ("units"), ("Hz",))

        # NXsource: /entry/instrument/source
        source = self.beamline.source
        nxsource = nxinstr.create_group("source")
        create_attributes(nxsource, ("NX_class",), ("NXsource",))
        nxsource.create_dataset("name", data=numpy.string_(source.name))
        create_attributes(nxsource["name"], ("short_name",), (source.short_name,))
        nxsource.create_dataset("type", data=numpy.string_(source.type))

        # NXdetector: /entry/instrument/detector
        self.write_NXdetector(nxinstr)

        # NXpositioner: /entry/instrument/detector_z, /entry/instrument/transformations
        self.write_NXpositioner(nxinstr)

    def write_NXsample(self, nxentry):
        goniometer = self.goniometer
        nxsample = nxentry.create_group("sample")
        create_attributes(nxsample, ("NX_class",), ("NXsample",))
        # Sample depends_on
        nxsample.create_dataset(
            "depends_on",
            data=set_dependency(
                goniometer.axes[-1], path="/entry/sample/transformations/"
            ),
        )
        # Beam: /entry/sample/beam
        nxsample["beam"] = self._nxs["/entry/instrument/beam"]

        # Transformations: /entry/sample/transformations
        nxtr = nxsample.create_group("transformations")
        create_attributes(nxtr, ("NX_class",), ("NXtransformations",))

        # Get a list of axis vectors
        vectors = NexusWriter.split_arrays(
            self._cf, goniometer.axes, goniometer.vectors
        )

        # Axes: /entry/sample/sample_
        # idx = numpy.where(goniometer.starts != goniometer.ends)[0][0]
        idx = [(i != j) for i, j in zip(goniometer.starts, goniometer.ends)].index(
            True
        )  # TEMPORARY FIX p2
        scan_axis = goniometer.axes[idx]
        for k in range(len(goniometer.axes)):
            if goniometer.axes[k] == scan_axis:
                nxscan = nxsample.create_group("sample_" + scan_axis)
                create_attributes(nxscan, ("NX_class",), ("NXpositioner",))
                nxscan[scan_axis] = self._nxs["/entry/data/" + scan_axis]
                nxtr[scan_axis] = self._nxs[
                    "/entry/sample/sample_" + scan_axis + "/" + scan_axis
                ]
            else:
                if "sam" in goniometer.axes[k]:
                    grp = "sample_" + goniometer.axes[k].split("_")[1]
                else:
                    grp = "sample_" + goniometer.axes[k]
                nxax = nxsample.create_group(grp)
                create_attributes(nxax, ("NX_class",), ("NXpositioner",))
                ax = nxax.create_dataset(goniometer.axes[k], data=goniometer.starts[k])
                _dep = set_dependency(
                    goniometer.depends[k], path="/entry/sample/transformations/"
                )
                create_attributes(
                    ax,
                    ("depends_on", "transformation_type", "units", "vector"),
                    (
                        _dep,
                        goniometer.types[k],
                        goniometer.units[k],
                        vectors[k],
                    ),
                )
                nxtr[goniometer.axes[k]] = self._nxs[
                    "/entry/sample/" + grp + "/" + goniometer.axes[k]
                ]

    def write(self):
        # Start writing the NeXus tree
        nxentry = self._nxs.create_group("entry")
        create_attributes(nxentry, ("NX_class",), ("NXentry",))

        # Definition: /entry/definition
        nxentry.create_dataset("definition", data=numpy.string_("NXmx"))

        # Data: /entry/data
        self.write_NXdata(nxentry)

        # Instrument: /entry/instrument
        self.write_NXinstrument(nxentry)

        # Sample: /entry/sample
        self.write_NXsample(nxentry)

        # Start and end time: /entry/start_time, /entry/end_time
        # Type: NX_DATE_TIME.
        # Start_time: time of collection first data point
        # end_time: time of collection last data point
        nxentry.create_dataset(
            "start_time", data=numpy.string_(self._params.input.start_time)
        )
        nxentry.create_dataset(
            "end_time", data=numpy.string_(self._params.input.end_time)
        )


"""
if __name__ == "__main__":
    import sys
    params = sys.argv[1:]
    with h5py.File(args.nxs_file, "x") as fh:
        NexusWriter(fh, params).write()
"""



"""
Utilities for creating NeXus format files.
"""

__author__ = "Diamond Light Source - Scientific Software"
__email__ = "scientificsoftware@diamond.ac.uk"
__version__ = "0.4.6"
__version_tuple__ = tuple(int(x) for x in __version__.split("."))

import numpy as np

from h5py import AttributeManager


def imgcif2mcstas(vector):
    """
    Convert from the standard coordinate frame used by imgCIF/CBF to the
    NeXus McStas coordinate system.

    Args:
        vector: Array of coordinates
    Returns:
        Tuple with the converted coordinate values
    """
    c2n = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, -1]])
    return tuple(np.dot(c2n, vector))


def create_attributes(nxs_obj, names, values):
    """
    Create or overwrite attributes with additional metadata information.

    Args:
        nxs_obj: NeXus object (Group or Dataset) to which the attributes should be attached
        names: Tuple containing the names of the new attributes
        values: Tuple containing the values relative to the names
    """
    for n, v in zip(names, values):
        if type(v) is str:
            # If a string, convert to numpy.string_
            v = np.string_(v)
        AttributeManager.create(nxs_obj, name=n, data=v)


def set_dependency(dep_info, path=None):
    """
    Define value for "depends_on" attribute.
    If the attribute points to the head of the dependency chain, simply pass "." for dep_info.

    Args:
        dep_info: The name of the transformation upon which the current one depends on.
        path: Where the transformation is. Set to None, if passed it points to location in the NeXus tree.
    Returns:
        The value to be passed to the attribute "depends_on"
    """
    if dep_info == ".":
        return np.string_(".")
    #    _d = dep_info
    #    if _d == ".":
    #        return numpy.string_(_d)
    #    else:
    #        _s = path + _d
    #        return numpy.string_(_s)

    if path:
        return np.string_(path + dep_info)
    else:
        return np.string_(dep_info)



"""
Tools for copying the metadata from Tristan NeXus files.
"""

import os
import h5py
import numpy as np

# from pathlib import Path

from . import get_nexus_tree, identify_scan_axis, convert_scan_axis
from .. import create_attributes


def single_image_nexus(data_file, tristan_nexus, write_mode="x"):
    """
    Create a NeXus file for a single-image data set.

    Copy the nexus tree from the original NeXus file for a collection on Tristan
    detector. In this case the input scan_axis is a tuple with the same start and
    stop value. The scan_axis in the new file will therefore be one single value.

    Args:
        data_file:      HDF5 file containing the newly binned images.
        tristan_nexus:  NeXus file with experiment metadata to be copied.
        write_mode:     Mode for writing the output NeXus file.  Accepts any valid
                        h5py file opening mode.

    Returns:
        The name of the output NeXus file.
    """
    nxs_filename = os.path.splitext(data_file)[0] + ".nxs"
    with h5py.File(tristan_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, write_mode
    ) as nxs_out:
        # Copy the whole tree except for nxdata
        nxentry = get_nexus_tree(nxs_in, nxs_out)
        # Create nxdata group
        nxdata = nxentry.create_group("data")
        # Add link to data
        nxdata["data"] = h5py.ExternalLink(data_file, "data")
        # Compute and write axis information
        ax, ax_attr = identify_scan_axis(nxs_in)
        create_attributes(
            nxdata, ("NX_class", "axes", "signal"), ("NXdata", ax, "data")
        )
        ax_range = nxs_in["entry/data"][ax][0]
        nxdata.create_dataset(ax, data=ax_range)
        # Write the attributes
        for key, value in ax_attr:
            nxdata[ax].attrs.create(key, value)
        # Now fix all other instances of scan_axis in the tree
        nxsample = nxentry["sample"]
        convert_scan_axis(nxsample, nxdata, ax)

    return nxs_filename


def multiple_images_nexus(
    data_file, tristan_nexus, write_mode="x", ang_vel=None, nbins=None
):
    """
    Create a NeXus file for a multiple-image data set.

    Copy the nexus tree from the original NeXus file for a collection on Tristan
    detector. In this case multiple images from a rotation collection have been
    binned. Thus the scan_axis in the input file is a tuple (start, stop). The
    scan_axis in the new file will therefore be a list of angles. ang_vel and
    num_bins are mutually exclusive arguments to work out the scan_axis list.

    Args:
        data_file:      HDF5 file containing the newly binned images.
        tristan_nexus:  NeXus file with experiment metadata to be copied.
        write_mode:     Mode for writing the output NeXus file.  Accepts any valid
                        h5py file opening mode.
        ang_vel:        Angular velocity, deg/s.
        nbins:          Number of binned images.

    Returns:
        The name of the output NeXus file.
    """
    nxs_filename = os.path.splitext(data_file)[0] + ".nxs"
    with h5py.File(tristan_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, write_mode
    ) as nxs_out:
        # Copy the whole tree except for nxdata
        nxentry = get_nexus_tree(nxs_in, nxs_out)
        # Create nxdata group
        nxdata = nxentry.create_group("data")
        # Add link to data
        nxdata["data"] = h5py.ExternalLink(data_file, "data")
        # Compute and write axis information
        ax, ax_attr = identify_scan_axis(nxs_in)
        create_attributes(
            nxdata, ("NX_class", "axes", "signal"), ("NXdata", ax, "data")
        )
        (start, stop) = nxs_in["entry/data"][ax][()]

        if ang_vel and nbins:
            raise ValueError(
                "ang_vel and nbins are mutually exclusive, "
                "please pass only one of them."
            )
        elif ang_vel:
            ax_range = np.array([round(p, 1) for p in np.arange(start, stop, ang_vel)])
        elif nbins:
            step = round(abs((stop - start) / nbins), 2)
            ax_range = np.array([round(p, 1) for p in np.arange(start, stop, step)])
        else:
            raise ValueError(
                "Impossible to calculate scan_axis, "
                "please pass either ang_vel or nbins."
            )

        nxdata.create_dataset(ax, data=ax_range)
        # Write the attributes
        for key, value in ax_attr:
            nxdata[ax].attrs.create(key, value)
        # Now fix all other instances of scan_axis in the tree
        nxsample = nxentry["sample"]
        convert_scan_axis(nxsample, nxdata, ax)

    return nxs_filename


def pump_probe_nexus(data_file, tristan_nexus, write_mode="x", mode="static"):
    """
    Create a NeXus file for a pump-probe image data set.

    Copy the nexus tree from the original NeXus file for a collection on Tristan
    detector. In this case multiple images from a pump-probe experiment have been
    binned. Thus the scan_axis in the input file is a tuple (start, stop). The
    scan_axis in the new file will be a single value if chosen mode is not
    "rotation". TBD rotation + pump-probe

    Args:
        data_file:      HDF5 file containing the newly binned images.
        tristan_nexus:  NeXus file with experiment metadata to be copied.
        write_mode:     Mode for writing the output NeXus file.  Accepts any valid
                        h5py file opening mode.
        mode:           "static", "powder_diffraction" or "rotation"

    Returns:
        The name of the output NeXus file.
    """
    # TODO: figure out a better way
    assert mode in ["static", "powder_diffraction", "rotation"], (
        "Mode passed is not valid, please pass one of the following "
        "['static', 'powder_diffraction', 'rotation']"
    )

    nxs_filename = os.path.splitext(data_file)[0] + ".nxs"
    with h5py.File(tristan_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, write_mode
    ) as nxs_out:
        # Copy the whole tree except for nxdata
        nxentry = get_nexus_tree(nxs_in, nxs_out)
        # Create nxdata group
        nxdata = nxentry.create_group("data")
        # Add link to data
        nxdata["data"] = h5py.ExternalLink(data_file, "data")
        # Compute and write axis information
        ax, ax_attr = identify_scan_axis(nxs_in)
        create_attributes(
            nxdata, ("NX_class", "axes", "signal"), ("NXdata", ax, "data")
        )
        if mode in ["static", "powder_diffraction"]:
            ax_range = nxs_in["entry/data"][ax][0]
        # else:
        # TODO
        nxdata.create_dataset(ax, data=ax_range)
        # Write the attributes
        for key, value in ax_attr:
            nxdata[ax].attrs.create(key, value)
        # Now fix all other instances of scan_axis in the tree
        nxsample = nxentry["sample"]
        convert_scan_axis(nxsample, nxdata, ax)

    return nxs_filename



"""
Utilities for copying metadata to new NeXus files.
"""

import h5py

from .. import create_attributes


def get_nexus_tree(nxs_in: h5py.File, nxs_out: h5py.File, skip=True):
    """
    Copy the tree from the original NeXus file, except for NXdata.

    Args:
        nxs_in: original NeXus file.
        nxs_out:new NeXus file.
        skip:   default True, copy everything but NXdata.
                Pass False to copy also NXdata.
    Returns:
        nxentry field
        nothing is skip is False
    """
    if skip is True:
        nxentry = nxs_out.create_group("entry")
        create_attributes(nxentry, ("NX_class",), ("NXentry",))
        # Copy all of the nexus tree as it is except for /entry/data
        for k in nxs_in["entry"].keys():
            if k == "data":
                continue
            nxs_in["entry"].copy(k, nxentry)
        return nxentry
    else:
        # Then copy everything, even data
        nxs_in.copy("entry", nxs_out)
        return


def identify_scan_axis(nxs_in: h5py.File):
    """
    Identify the scan_axis in the NeXus tree.

    Args:
        nxs_in: Tristan NeXus file
    Returns:
        ax:         Name of the scan_axis
        ax_attrs:   Attributes of the scan_axis dataset, type h5py._hl.base.ItemsViewHDF5
    """
    nxdata = nxs_in["entry/data"]
    for k in nxdata.keys():
        try:
            if type(nxdata[k]) is h5py.Dataset:
                # in Tristan NeXus files everything else is a group
                ax = k
                ax_attr = nxdata[k].attrs.items()
        except KeyError:
            continue
    return ax, ax_attr


def convert_scan_axis(nxsample, nxdata, ax):
    """
    Modify all instances of scan_axis present in NeXus file NXsample group.

    Args:
        nxsample:   NXsample group of NeXus file to be modified
        nxdata:     NXdata group of NeXus file to be modified
        ax:         Name of scan_axis
    """
    del nxsample["transformations/" + ax]
    nxsample["transformations/" + ax] = nxdata[ax]
    name = "sample_" + ax + "/" + ax
    del nxsample[name]
    nxsample[name] = nxdata[ax]



"""
General tools to copy metadata from NeXus files.
"""
import os
import h5py

from . import get_nexus_tree
from .. import create_attributes


def images_nexus(data_file, original_nexus, simple_copy=False):
    """
    Copy NeXus metadata for images.

    Args:
        data_file:       HDF5 file.
        original_nexus:  NeXus file with experiment metadata.
        simple_copy:     Default False. If passed, copy everything directly.
    """
    nxs_filename = os.path.splitext(data_file)[0] + ".nxs"
    with h5py.File(original_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, "x"
    ) as nxs_out:
        if simple_copy is True:
            # Copy the whole tree including nxdata
            get_nexus_tree(nxs_in, nxs_out, skip=False)
        else:
            # Copy the whole tree except for nxdata
            nxentry = get_nexus_tree(nxs_in, nxs_out)
            # Create nxdata group
            nxdata = nxentry.create_group("data")
            # Find and copy only scan axis information
            for k in nxs_in["entry/data"].keys():
                if "depends_on" in nxs_in["entry/data"][k].attrs.keys():
                    ax = k
                    nxs_in["entry/data"].copy(k, nxdata)
            create_attributes(
                nxdata, ("NX_class", "axes", "signal"), ("NXdata", ax, "data")
            )
            # Add link to data
            with h5py.File(data_file, "r") as fout:
                nxdata["data"] = h5py.ExternalLink(fout.filename, "data")


def pseudo_events_nexus(data_file, original_nexus):
    """
    Copy NeXus metadata for pseudo event mode data.

    Args:
        data_file:       HDF5 file.
        original_nexus:  NeXus file with experiment metadata.
    """
    nxs_filename = os.path.splitext(data_file)[0] + ".nxs"
    with h5py.File(original_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, "x"
    ) as nxs_out:
        # Copy the whole tree except for nxdata
        nxentry = get_nexus_tree(nxs_in, nxs_out)
        # Create nxdata group
        nxdata = nxentry.create_group("data")
        # Find and copy only scan axis information
        for k in nxs_in["entry/data"].keys():
            if "depends_on" in nxs_in["entry/data"][k].attrs.keys():
                ax = k
                nxs_in["entry/data"].copy(k, nxdata)
        create_attributes(
            nxdata, ("NX_class", "axes", "signal"), ("NXdata", ax, "data")
        )
        # Add link to data
        with h5py.File(data_file, "r") as fout:
            for k in fout.keys():
                nxdata[k] = h5py.ExternalLink(fout.filename, k)



import freephil

# scope for the geometry input - N.B. having to unroll the input for structured
# axis definitions as PHIL does not work well with complex data structures ->
# have axis names, axis values, ... as separate parameters. Ah, this already
# exists as a goniometer phil scope in dxtbx. Will have to assume that the
# axes are presented in order... ah, no, because we need translation axes in
# the stack as well.

scope = freephil.parse(
    """
output {
  data_file_template = nexus_%06d.h5
    .type = path
    .help = "Filename template for data files"
  master_file_name = nexus_master.h5
    .type = path
    .help = "Filename for master file"
}
input {
  coordinate_frame = mcstas *imgcif
    .type = choice
    .help = "Which coordinate system is being used to provide input vectors"
  n_images = 2
    .type = int
    .help = "Number of blank images to be generated per file"
  n_events = 100000
    .type = int
    .help = "Length of event stream per file"
  detector {
    mode = *images events
      .type = choice
      .help = "Which kind of data is generated by the detector."
    description = Eiger 2XE 16M
      .type = str
      .help = "Detector class to record"
    sensor_material = *Si CdTe
      .type = choice
      .help = "Sensor material (e.g. silicon)"
    sensor_thickness = 0.320
      .type = float
      .help = "Sensor thickness, mm"
    overload = 65535
      .type = int
      .help = "Pixels >= this value are invalid due to overloading"
    underload = -1
      .type = int
      .help = "Pixels <= this value are invalid"
    pixel_size = 0.075 0.075
      .type = floats(size = 2)
      .help = "Pixel size in mm"
    beam_center = 2214.355 2300.496
      .type = floats(size = 2)
      .help = "Beam position on the detector"
    flatfield = None
      .type = path
      .help = "If path is given, add flatfield correction data field"
    pixel_mask = None
      .type = path
      .help = "if path is given, add link to bad pixel mask"
# for the moment only considering single modules but will have to consider
# multiple modules in the not too distant future
    image_size = 4362 4148
      .type = ints(size = 2)
      .help = "Image size, slow, fast"
    exposure_time = 0.004
      .type = float
      .help = "Nominal exposure time, seconds"
# taken as a given that fast axis sits on slow axis which sits on last axis
# named below - should probably sort out the axis offset definition i.e. are
# the offsets applied before or after the axis shift?
    fast_axis = 1 0 0
      .type = floats(size = 3)
      .help = "Fast axis at datum position"
    slow_axis = 0 -1 0
      .type = floats(size = 3)
      .help = "Slow axis at datum position"
    offsets = -166.2 172.1 0 0 0 0
      .type = floats
      .help = "Axis offsets - one after the other - fast then slow"
    module_offset = False
      .type = bool
      .help = "If set to true, calculates offset of the module in regard to detector origin and creates corresponding field"
    axes = two_theta det_z
      .type = strings
      .help = "Axis names for the detector axes"
    depends = . two_theta
      .type = strings
      .help = "Axis names for the axis dependencies"
    vectors = 1 0 0 0 0 -1
      .type = floats
      .help = "Axis vectors - one after the other"
    starts = 0.0 100.0
      .type = floats
      .help = "Starting values for axes"
    ends = 0.0 100.0
      .type = floats
      .help = "Ending values for axes, only different to start for scan axis"
    increments = 0.0 0.0
      .type = floats
      .help = "Increment values for axes, non-zero only for scan axis"
    types = rotation translation
      .type = strings
      .help = "Axis types, rotation or translation"
    units = deg mm
      .type = strings
      .help = "Axis units, from mm or deg"
  }

  goniometer {
# sample is assumed to sit on the last of tehse axes
    axes = omega sam_z sam_y sam_x chi phi
      .type = strings
      .help = "Axis names for the goniometer axes"
    depends = . omega sam_z sam_y sam_x chi
      .type = strings
      .help = "Axis names for the axis dependencies"
    vectors = 1 0 0 0 0 1 0 1 0 1 0 0 0 0 1 1 0 0
      .type = floats
      .help = "Axis vectors - one after the other"
    offsets = 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
      .type = floats
      .help = "Axis offsets - one after the other"
    starts = 0.0 0.0 0.0 0.0 0.0 0.0
      .type = floats
      .help = "Starting values for axes"
    ends = 90.0 0.0 0.0 0.0 0.0 0.0
      .type = floats
      .help = "Ending values for axes, only different to start for scan axis"
    increments = 0.1 0.0 0.0 0.0 0.0 0.0
      .type = floats
      .help = "Increment values for axes, non-zero only for scan axis"
    types = rotation translation translation translation rotation rotation
      .type = strings
      .help = "Axis types, rotation or translation"
    units = deg mm mm mm deg deg deg
      .type = strings
      .help = "Axis units, from mm or deg"
  }

  beamline {
    source {
      name = Diamond Light Source
        .type = str
        .help = "Facility name"
      short_name = DLS
        .type = str
        .help = "Facility abbreviation"
      type = Synchrotron X-ray Source
        .type = str
        .help = "Facility type"
    }
  }
  beam {
    wavelength = 0.979590
      .type = float
      .help = "Wavelength of incident beam, angstroms"
    flux = 268717230611.358
      .type = float
      .help = "Flux of incident beam, ph / s"
  }
  attenuator {
    transmission = 1
      .type = float
      .help = "Attenuation of beam intensity"
  }

# these are ornamental since there was never an experiment
  start_time = Wed Nov 18 11:08:02 2020
    .type = str
    .help = "Experiment start time"
  end_time = Wed Nov 18 11:08:02 2020
    .type = str
    .help = "Experiment end time"
}
"""
)

if __name__ == "__main__":
    print(scope.as_str())
